# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from minio import Minio
from minio.select import (CSVInputSerialization, CSVOutputSerialization,
                          SelectRequest)

client = Minio(
    endpoint="play.min.io",
    access_key="Q3AM3UQ867SPQQA43P2F",
    secret_key="zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG",
)

with client.select_object_content(
        bucket_name="my-bucket",
        object_name="my-object.csv",
        request=SelectRequest(
            expression="select * from S3Object",
            input_serialization=CSVInputSerialization(),
            output_serialization=CSVOutputSerialization(),
            request_progress=True,
        ),
) as result:
    for data in result.stream():
        print(data.decode())
    print(result.stats())



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2024 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import locale
import threading
import unittest
from contextlib import contextmanager
from datetime import datetime, timezone
from typing import Any, Generator

from minio.time import from_http_header, to_http_header

LOCALE_LOCK = threading.Lock()
LAST_MODIFIED_STR = "Mon, 02 Mar 2015 07:28:00 GMT"
LAST_MODIFIED_DATE = datetime(
    year=2015,
    month=3,
    day=2,
    hour=7,
    minute=28,
    second=0,
    tzinfo=timezone.utc,
)


@contextmanager
def setlocale(name) -> Generator[str, Any, None]:
    with LOCALE_LOCK:
        saved = locale.setlocale(locale.LC_ALL)
        try:
            yield locale.setlocale(locale.LC_ALL, name)
        finally:
            locale.setlocale(locale.LC_ALL, saved)


class TimeutilsTest(unittest.TestCase):
    def test_from_http_header_valid_headers(self) -> None:
        for case in [
            (
                "Wed, 30 Oct 2024 09:35:00 GMT",
                datetime(
                    year=2024,
                    month=10,
                    day=30,
                    hour=9,
                    minute=35,
                    second=0,
                    tzinfo=timezone.utc,
                ),
            ),
            (
                "Tue, 29 Oct 2024 00:35:00 GMT",
                datetime(
                    year=2024,
                    month=10,
                    day=29,
                    hour=0,
                    minute=35,
                    second=0,
                    tzinfo=timezone.utc,
                ),
            ),
            (
                "Tue, 01 Oct 2024 22:35:22 GMT",
                datetime(
                    year=2024,
                    month=10,
                    day=1,
                    hour=22,
                    minute=35,
                    second=22,
                    tzinfo=timezone.utc,
                ),
            ),
            (
                "Mon, 30 Sep 2024 22:35:55 GMT",
                datetime(
                    year=2024,
                    month=9,
                    day=30,
                    hour=22,
                    minute=35,
                    second=55,
                    tzinfo=timezone.utc,
                ),
            ),
        ]:
            with self.subTest(case=case):
                self.assertEqual(from_http_header(case[0]), case[1])

    def test_from_http_header_invalid_headers(self) -> None:
        for case in [
            ("Wed, 30 Oct 2024 09:35:00 GMT ", "invalid length"),
            ("Wet, 30 Oct 2024 09:35:00 GMT", "invalid weekday"),
            ("Wed  30 Oct 2024 09:35:00 GMT", "no comma after weekday"),
            ("Wed,30 Sep 2024 09:35:00 GMT ", "no space after weekday"),
            ("Wed, 30Sep 2024 09:35:00 GMT ", "no space before month"),
            ("Wed, 00 Sep 2024 09:35:00 GMT", "invalid day"),
            ("Wed, 32 Sep 2024 09:35:00 GMT", "invalid day 2"),
            ("Wed, ab Sep 2024 09:35:00 GMT", "invalid day 3"),
            ("Wed, 30 Set 2024 09:35:00 GMT", "invalid month"),
            ("Tue, 30 Set 2024 09:35:00 GMT", "name of day doesn't match"),
        ]:
            with self.subTest(case=case):
                self.assertRaises(ValueError, from_http_header, case[0])

    def test_from_http_header_default_locale(self) -> None:
        result_datetime = from_http_header(LAST_MODIFIED_STR)

        self.assertEqual(
            result_datetime,
            LAST_MODIFIED_DATE,
        )

    def test_from_http_header_polish_locale(self) -> None:
        try:
            with setlocale("pl_PL.utf8"):
                result_datetime = from_http_header(LAST_MODIFIED_STR)

                self.assertEqual(
                    result_datetime,
                    LAST_MODIFIED_DATE,
                )
        except locale.Error:
            self.skipTest("pl_PL.utf8 locale is not supported on this machine")

    def test_to_http_header_default_locale(self) -> None:
        self.assertEqual(
            to_http_header(LAST_MODIFIED_DATE),
            LAST_MODIFIED_STR,
        )

    def test_to_http_header_polish_locale(self) -> None:
        try:
            with setlocale("pl_PL.utf8"):
                self.assertEqual(
                    to_http_header(LAST_MODIFIED_DATE),
                    LAST_MODIFIED_STR,
                )
        except locale.Error:
            self.skipTest("pl_PL.utf8 locale is not supported on this machine")



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import TestCase

from minio.credentials.providers import (AWSConfigProvider, ChainedProvider,
                                         EnvAWSProvider, EnvMinioProvider,
                                         MinioClientConfigProvider,
                                         StaticProvider)

CONFIG_JSON_SAMPLE = "tests/unit/config.json.sample"
CREDENTIALS_SAMPLE = "tests/unit/credentials.sample"
CREDENTIALS_EMPTY = "tests/unit/credentials.empty"


class CredentialsTest(TestCase):
    def test_credentials_get(self):
        provider = MinioClientConfigProvider(
            filename=CONFIG_JSON_SAMPLE,
            alias="play",
        )
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "Q3AM3UQ867SPQQA43P2F")
        self.assertEqual(creds.secret_key,
                         "zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG")
        self.assertEqual(creds.session_token, None)


class ChainedProviderTest(TestCase):
    def test_chain_retrieve(self):
        # clear environment
        os.environ.clear()
        # prepare env for env_aws provider
        os.environ["AWS_ACCESS_KEY_ID"] = "access_aws"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "secret_aws"
        os.environ["AWS_SESSION_TOKEN"] = "token_aws"
        # prepare env for env_minio
        os.environ["MINIO_ACCESS_KEY"] = "access_minio"
        os.environ["MINIO_SECRET_KEY"] = "secret_minio"
        # create chain provider with env_aws and env_minio providers

        provider = ChainedProvider(
            [
                EnvAWSProvider(), EnvMinioProvider(),
            ]
        )
        # retrieve provider (env_aws) has priority
        creds = provider.retrieve()
        # assert provider credentials
        self.assertEqual(creds.access_key, "access_aws")
        self.assertEqual(creds.secret_key, "secret_aws")
        self.assertEqual(creds.session_token, "token_aws")

    def test_chain_retrieve_failed_provider(self):
        # clear environment
        os.environ.clear()
        # prepare env for env_minio
        os.environ["MINIO_ACCESS_KEY"] = "access_minio"
        os.environ["MINIO_SECRET_KEY"] = "secret_minio"
        # create chain provider with env_aws and env_minio providers

        provider = ChainedProvider(
            [
                EnvAWSProvider(), EnvMinioProvider(),
            ]
        )
        # retrieve provider: (env_minio) will be retrieved
        creds = provider.retrieve()
        # assert provider credentials
        self.assertEqual(creds.access_key, "access_minio")
        self.assertEqual(creds.secret_key, "secret_minio")
        self.assertEqual(creds.session_token, None)


class EnvAWSProviderTest(TestCase):
    def test_env_aws_retrieve(self):
        os.environ.clear()
        os.environ["AWS_ACCESS_KEY_ID"] = "access"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "secret"
        os.environ["AWS_SESSION_TOKEN"] = "token"
        provider = EnvAWSProvider()
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "access")
        self.assertEqual(creds.secret_key, "secret")
        self.assertEqual(creds.session_token, "token")

    def test_env_aws_retrieve_no_token(self):
        os.environ.clear()
        os.environ["AWS_ACCESS_KEY_ID"] = "access"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "secret"
        provider = EnvAWSProvider()
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "access")
        self.assertEqual(creds.secret_key, "secret")
        self.assertEqual(creds.session_token, None)


class EnvMinioTest(TestCase):
    def test_env_minio_retrieve(self):
        os.environ.clear()
        os.environ['MINIO_ACCESS_KEY'] = "access"
        os.environ["MINIO_SECRET_KEY"] = "secret"
        provider = EnvMinioProvider()
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "access")
        self.assertEqual(creds.secret_key, "secret")
        self.assertEqual(creds.session_token, None)


class AWSConfigProviderTest(TestCase):
    def test_file_aws(self):
        os.environ.clear()
        provider = AWSConfigProvider(CREDENTIALS_SAMPLE)
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "accessKey")
        self.assertEqual(creds.secret_key, "secret")
        self.assertEqual(creds.session_token, "token")

    def test_file_aws_from_env(self):
        os.environ.clear()
        os.environ["AWS_SHARED_CREDENTIALS_FILE"] = (
            CREDENTIALS_SAMPLE
        )
        provider = AWSConfigProvider()
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "accessKey")
        self.assertEqual(creds.secret_key, "secret")
        self.assertEqual(creds.session_token, "token")

    def test_file_aws_env_profile(self):
        os.environ.clear()
        os.environ["AWS_PROFILE"] = "no_token"
        provider = AWSConfigProvider(CREDENTIALS_SAMPLE)
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "accessKey")
        self.assertEqual(creds.secret_key, "secret")
        self.assertEqual(creds.session_token, None)

    def test_file_aws_arg_profile(self):
        os.environ.clear()
        provider = AWSConfigProvider(
            CREDENTIALS_SAMPLE,
            "no_token",
        )
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "accessKey")
        self.assertEqual(creds.secret_key, "secret")
        self.assertEqual(creds.session_token, None)

    def test_file_aws_no_creds(self):
        os.environ.clear()
        provider = AWSConfigProvider(
            CREDENTIALS_EMPTY,
            "no_token",
        )
        try:
            provider.retrieve()
        except ValueError:
            pass


class MinioClientConfigProviderTest(TestCase):
    def test_file_minio_(self):
        os.environ.clear()
        provider = MinioClientConfigProvider(filename=CONFIG_JSON_SAMPLE)
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "accessKey")
        self.assertEqual(creds.secret_key, "secret")
        self.assertEqual(creds.session_token, None)

    def test_file_minio_env_alias(self):
        os.environ.clear()
        os.environ["MINIO_ALIAS"] = "play"
        provider = MinioClientConfigProvider(filename=CONFIG_JSON_SAMPLE)
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "Q3AM3UQ867SPQQA43P2F")
        self.assertEqual(creds.secret_key,
                         "zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG")
        self.assertEqual(creds.session_token, None)

    def test_file_minio_arg_alias(self):
        os.environ.clear()
        provider = MinioClientConfigProvider(
            filename=CONFIG_JSON_SAMPLE,
            alias="play",
        )
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "Q3AM3UQ867SPQQA43P2F")
        self.assertEqual(creds.secret_key,
                         "zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG")
        self.assertEqual(creds.session_token, None)


class StaticProviderTest(TestCase):
    def test_static_credentials(self):
        provider = StaticProvider("UXHW", "SECRET")
        creds = provider.retrieve()
        self.assertEqual(creds.access_key, "UXHW")
        self.assertEqual(creds.secret_key, "SECRET")
        self.assertEqual(creds.session_token, None)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2025 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from urllib.request import urlopen

from minio import Minio

client = Minio(
    endpoint="play.min.io",
    access_key="Q3AM3UQ867SPQQA43P2F",
    secret_key="zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG",
)

# Upload data.
result = client.put_object(
    bucket_name="my-bucket",
    object_name="my-object",
    data=io.BytesIO(b"hello, "),
    length=7,
)
print(f"created {result.object_name} object; etag: {result.etag}")

# Append data.
result = client.append_object(
    bucket_name="my-bucket",
    object_name="my-object",
    data=io.BytesIO(b"world"),
    length=5,
)
print(f"appended {result.object_name} object; etag: {result.etag}")

# Append data in chunks.
with urlopen(
    "https://www.kernel.org/pub/linux/kernel/v6.x/linux-6.13.12.tar.xz",
) as stream:
    result = client.append_object(
        bucket_name="my-bucket",
        object_name="my-object",
        stream=stream,
        length=148611164,
        chunk_size=5*1024*1024,
    )
    print(f"appended {result.object_name} object; etag: {result.etag}")

# Append unknown sized data.
with urlopen(
    "https://www.kernel.org/pub/linux/kernel/v6.x/linux-6.14.3.tar.xz",
) as stream:
    result = client.append_object(
        bucket_name="my-bucket",
        object_name="my-object",
        stream=stream,
        chunk_size=5*1024*1024,
    )
    print(f"appended {result.object_name} object; etag: {result.etag}")



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import hmac
from datetime import datetime, timezone
from unittest import TestCase
from urllib.parse import urlsplit, urlunsplit

from urllib3._collections import HTTPHeaderDict

from minio import Minio
from minio.credentials import Credentials
from minio.helpers import queryencode, quote, sha256_hash
from minio.signer import (_get_authorization, _get_canonical_request_hash,
                          _get_scope, _get_signing_key, _get_string_to_sign,
                          presign_v4, sign_v4_s3)

empty_hash = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
dt = datetime(2015, 6, 20, 1, 2, 3, 0, timezone.utc)


class CanonicalRequestTest(TestCase):
    def test_simple_request(self):
        url = urlsplit('http://localhost:9000/hello')
        expected_signed_headers = ['x-amz-content-sha256', 'x-amz-date']
        expected_request_array = ['PUT', '/hello', '',
                                  'x-amz-content-sha256:' +
                                  empty_hash, 'x-amz-date:dateString',
                                  '', ';'.join(expected_signed_headers),
                                  empty_hash]
        headers_to_sign = HTTPHeaderDict(
            {'x-amz-date': 'dateString', 'x-amz-content-sha256': empty_hash},
        )

        expected_request = sha256_hash('\n'.join(expected_request_array))
        actual_request = _get_canonical_request_hash(
            method="PUT",
            url=url,
            headers=headers_to_sign,
            content_sha256=empty_hash,
        )
        self.assertEqual(expected_request, actual_request[0])

    def test_request_with_query(self):
        url = urlsplit('http://localhost:9000/hello?c=d&e=f&a=b')
        expected_signed_headers = ['x-amz-content-sha256', 'x-amz-date']
        expected_request_array = ['PUT', '/hello', 'a=b&c=d&e=f',
                                  'x-amz-content-sha256:' + empty_hash,
                                  'x-amz-date:dateString',
                                  '', ';'.join(expected_signed_headers),
                                  empty_hash]

        expected_request = sha256_hash('\n'.join(expected_request_array))

        headers_to_sign = HTTPHeaderDict(
            {'x-amz-date': 'dateString', 'x-amz-content-sha256': empty_hash},
        )
        actual_request = _get_canonical_request_hash(
            method="PUT",
            url=url,
            headers=headers_to_sign,
            content_sha256=empty_hash,
        )
        self.assertEqual(expected_request, actual_request[0])


class StringToSignTest(TestCase):
    def test_signing_key(self):
        expected_signing_key_list = [
            'AWS4-HMAC-SHA256', '20150620T010203Z',
            '20150620/us-east-1/s3/aws4_request',
            'b93e86965c269a0dfef37a8bec231ef8acf8cdb101a64eb700a46c452c1ad233'
        ]

        actual_signing_key = _get_string_to_sign(
            dt, _get_scope(dt, 'us-east-1', "s3"),
            'b93e86965c269a0dfef37a8bec231ef8acf8cdb101a64eb700a46c452c1ad233')
        self.assertEqual('\n'.join(expected_signing_key_list),
                         actual_signing_key)


class SigningKeyTest(TestCase):
    def test_generate_signing_key(self):
        key1_string = 'AWS4' + 'S3CR3T'
        key1 = key1_string.encode('utf-8')
        key2 = hmac.new(key1, '20150620'.encode(
            'utf-8'), hashlib.sha256).digest()
        key3 = hmac.new(key2, 'region'.encode(
            'utf-8'), hashlib.sha256).digest()
        key4 = hmac.new(key3, 's3'.encode('utf-8'), hashlib.sha256).digest()
        expected_result = hmac.new(key4, 'aws4_request'.encode(
            'utf-8'), hashlib.sha256).digest()

        actual_result = _get_signing_key('S3CR3T', dt, 'region', "s3")
        self.assertEqual(expected_result, actual_result)


class AuthorizationHeaderTest(TestCase):
    def test_generate_authentication_header(self):
        expected_authorization_header = (
            'AWS4-HMAC-SHA256 Credential='
            'public_key/20150620/region/s3/aws4_request, '
            'SignedHeaders=host;X-Amz-Content-Sha256;X-Amz-Date, '
            'Signature=signed_request'
        )
        actual_authorization_header = _get_authorization(
            'public_key', _get_scope(dt, 'region', "s3"),
            'host;X-Amz-Content-Sha256;X-Amz-Date', 'signed_request')
        self.assertEqual(expected_authorization_header,
                         actual_authorization_header)


class PresignURLTest(TestCase):
    def test_presigned_versioned_id(self):
        credentials = Credentials("minio", "minio123")
        url = presign_v4(
            method='GET',
            url=urlsplit(
                'http://localhost:9000/bucket-name/objectName?versionId=uuid'
            ),
            region='us-east-1',
            credentials=credentials,
            date=dt,
            expires=604800,
        )

        self.assertEqual(urlunsplit(url), 'http://localhost:9000/bucket-name/objectName?versionId=uuid&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=minio%2F20150620%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Date=20150620T010203Z&X-Amz-Expires=604800&X-Amz-SignedHeaders=host&X-Amz-Signature=3ce13e2ca929fafa20581a05730e4e9435f2a5e20ec7c5a082d175692fb0a663')


class SignV4Test(TestCase):
    def test_signv4(self):
        client = Minio(
            endpoint="localhost:9000",
            access_key="minio",
            secret_key="minio123",
            secure=False,
        )
        creds = client._provider.retrieve()
        headers = HTTPHeaderDict({
            'Host': 'localhost:9000',
            'x-amz-content-sha256':
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
            'x-amz-date': '20150620T010203Z',
        })
        url = client._base_url.build(
            method="PUT",
            region="us-east-1",
            bucket_name="testbucket",
            object_name="~testobject",
            query_params={"partID": "1", "uploadID": "~abcd"},
        )
        headers = sign_v4_s3(
            method="PUT",
            url=url,
            region="us-east-1",
            headers=headers,
            credentials=creds,
            content_sha256=(
                "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
            ),
            date=dt,
        )
        self.assertEqual(headers['Authorization'],
                         'AWS4-HMAC-SHA256 Credential='
                         'minio/20150620/us-east-1/s3/aws4_request, '
                         'SignedHeaders=host;x-amz-content-sha256;x-amz-date, '
                         'Signature='
                         'a2f4546f647981732bd90dfa5a7599c44dca92f44bea48ecc7565df06032c25b')


class UnicodeEncodeTest(TestCase):
    def test_unicode_quote(self):
        self.assertEqual(quote('/test/123/汉字'), '/test/123/%E6%B1%89%E5%AD%97')

    def test_unicode_queryencode(self):
        self.assertEqual(queryencode('/test/123/汉字'),
                         '%2Ftest%2F123%2F%E6%B1%89%E5%AD%97')

    def test_unicode_quote_u(self):
        self.assertEqual(quote(u'/test/123/汉字'),
                         '/test/123/%E6%B1%89%E5%AD%97')

    def test_unicode_queryencode_u(self):
        self.assertEqual(queryencode(u'/test/123/汉字'),
                         '%2Ftest%2F123%2F%E6%B1%89%E5%AD%97')

    def test_unicode_quote_b(self):
        self.assertEqual(quote(b'/test/123/\xe6\xb1\x89\xe5\xad\x97'),
                         '/test/123/%E6%B1%89%E5%AD%97')

    def test_unicode_queryencode_b(self):
        self.assertEqual(queryencode(b'/test/123/\xe6\xb1\x89\xe5\xad\x97'),
                         '%2Ftest%2F123%2F%E6%B1%89%E5%AD%97')



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# A Chain credentials provider, provides a way of chaining multiple providers
# together and will pick the first available using priority order of the
# 'providers' list

from minio import Minio
from minio.credentials import (AWSConfigProvider, ChainedProvider,
                               EnvAWSProvider, IamAwsProvider)

client = Minio(
    endpoint="s3.amazonaws.com",
    credentials=ChainedProvider(
        [
            IamAwsProvider(),
            AWSConfigProvider(),
            EnvAWSProvider(),
        ]
    )
)

# Get information of an object.
stat = client.stat_object(bucket_name="my-bucket", object_name="my-object")
print(stat)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2025 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Checksum functions."""

from __future__ import absolute_import, annotations, division, unicode_literals

import base64
import binascii
import hashlib
import struct
from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict, List, Optional

# MD5 hash of zero length byte array.
ZERO_MD5_HASH = "1B2M2Y8AsgTpgAmY7PhCfg=="
# SHA-256 hash of zero length byte array.
ZERO_SHA256_HASH = (
    "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
)
UNSIGNED_PAYLOAD = "UNSIGNED-PAYLOAD"


def base64_string(data: bytes) -> str:
    """Encodes the specified bytes to Base64 string."""
    return base64.b64encode(data).decode("ascii")


def base64_string_to_sum(value: str) -> bytes:
    """Decodes the specified Base64 encoded string to bytes."""
    return base64.b64decode(value)


def hex_string(data: bytes) -> str:
    """Encodes the specified bytes to Base16 (hex) string."""
    return "".join(f"{b:02x}" for b in data)


def hex_string_to_sum(value: str) -> bytes:
    """Decodes the specified Base16 (hex) encoded string to bytes."""
    if len(value) % 2 != 0:
        raise ValueError("Hex string length must be even")
    return bytes(int(value[i:i+2], 16) for i in range(0, len(value), 2))


class Hasher(ABC):
    """Checksum hasher interface."""

    @abstractmethod
    def update(
            self,
            data: bytes,
            offset: Optional[int] = None,
            length: Optional[int] = None,
    ) -> None:
        """Update the hash with bytes from b[off:off+length]."""

    @abstractmethod
    def sum(self) -> bytes:
        """Return the final digest."""

    @abstractmethod
    def reset(self) -> None:
        """Reset the hasher state."""


class CRC32(Hasher):
    """CRC32 Hasher using binascii.crc32."""

    def __init__(self):
        self._crc = 0

    def update(
            self,
            data: bytes,
            offset: Optional[int] = None,
            length: Optional[int] = None,
    ) -> None:
        offset = offset or 0
        if length is None:
            length = len(data) - offset
        self._crc = binascii.crc32(
            data[offset:offset+length], self._crc,
        ) & 0xFFFFFFFF

    def sum(self) -> bytes:
        return struct.pack(">I", self._crc)

    def reset(self) -> None:
        self._crc = 0


def _generate_crc32c_table():
    """Generates CRC32C table."""
    table = [0] * 256
    for i in range(256):
        crc = i
        for _ in range(8):
            crc = (crc >> 1) ^ (0x82F63B78 if (crc & 1) else 0)
        table[i] = crc & 0xFFFFFFFF
    return table


_CRC32C_TABLE = _generate_crc32c_table()


class CRC32C(Hasher):
    """CRC32C Hasher."""

    def __init__(self):
        self._crc = 0xFFFFFFFF

    def update(
            self,
            data: bytes,
            offset: Optional[int] = None,
            length: Optional[int] = None,
    ) -> None:
        offset = offset or 0
        if length is None:
            length = len(data) - offset
        for byte in data[offset:offset+length]:
            self._crc = _CRC32C_TABLE[
                (self._crc ^ byte) & 0xFF] ^ (self._crc >> 8)

    def sum(self) -> bytes:
        crc_final = (~self._crc) & 0xFFFFFFFF
        return crc_final.to_bytes(4, "big")

    def reset(self) -> None:
        self._crc = 0xFFFFFFFF


def _generate_crc64nvme_table():
    """Generates CRC64NVME table."""
    table = [0] * 256
    slicing8_table = [[0] * 256 for _ in range(8)]

    polynomial = 0x9A6C9329AC4BC9B5
    for i in range(256):
        crc = i
        for _ in range(8):
            if crc & 1:
                crc = (crc >> 1) ^ polynomial
            else:
                crc >>= 1
        table[i] = crc & 0xFFFFFFFFFFFFFFFF

    slicing8_table[0] = table[:]
    for i in range(256):
        crc = table[i]
        for j in range(1, 8):
            crc = table[crc & 0xFF] ^ (crc >> 8)
            slicing8_table[j][i] = crc & 0xFFFFFFFFFFFFFFFF

    return table, slicing8_table


_CRC64NVME_TABLE, _SLICING8_TABLE_NVME = _generate_crc64nvme_table()


class CRC64NVME(Hasher):
    """CRC64 NVME checksum."""

    def __init__(self):
        self._crc = 0

    def update(
            self,
            data: bytes,
            offset: Optional[int] = None,
            length: Optional[int] = None,
    ):
        offset = offset or 0
        if length is None:
            length = len(data) - offset
        data = data[offset:offset + length]
        self._crc = ~self._crc & 0xFFFFFFFFFFFFFFFF
        offset = 0

        # Process in 8-byte chunks (little-endian)
        while len(data) >= 64 and (len(data) - offset) > 8:
            value = struct.unpack_from("<Q", data, offset)[0]
            self._crc ^= value
            self._crc = (
                _SLICING8_TABLE_NVME[7][self._crc & 0xFF] ^
                _SLICING8_TABLE_NVME[6][(self._crc >> 8) & 0xFF] ^
                _SLICING8_TABLE_NVME[5][(self._crc >> 16) & 0xFF] ^
                _SLICING8_TABLE_NVME[4][(self._crc >> 24) & 0xFF] ^
                _SLICING8_TABLE_NVME[3][(self._crc >> 32) & 0xFF] ^
                _SLICING8_TABLE_NVME[2][(self._crc >> 40) & 0xFF] ^
                _SLICING8_TABLE_NVME[1][(self._crc >> 48) & 0xFF] ^
                _SLICING8_TABLE_NVME[0][(self._crc >> 56)]
            ) & 0xFFFFFFFFFFFFFFFF
            offset += 8

        # Process remaining bytes
        for i in range(offset, length):
            self._crc = (
                _CRC64NVME_TABLE[(self._crc ^ data[i]) & 0xFF] ^
                (self._crc >> 8)
            ) & 0xFFFFFFFFFFFFFFFF

        self._crc = ~self._crc & 0xFFFFFFFFFFFFFFFF

    def reset(self):
        self._crc = 0

    def sum(self) -> bytes:
        value = self._crc
        return bytes([
            (value >> 56) & 0xFF,
            (value >> 48) & 0xFF,
            (value >> 40) & 0xFF,
            (value >> 32) & 0xFF,
            (value >> 24) & 0xFF,
            (value >> 16) & 0xFF,
            (value >> 8) & 0xFF,
            value & 0xFF
        ])


class HashlibHasher(Hasher, ABC):
    """Generic wrapper for hashlib algorithms."""

    def __init__(self, name: str):
        self._name = name
        self._hasher = hashlib.new(name)

    def update(
            self,
            data: bytes,
            offset: Optional[int] = None,
            length: Optional[int] = None,
    ) -> None:
        offset = offset or 0
        if length is None:
            length = len(data) - offset
        self._hasher.update(data[offset:offset+length])

    def sum(self) -> bytes:
        return self._hasher.digest()

    def reset(self) -> None:
        self._hasher = hashlib.new(self._name)


class SHA1(HashlibHasher):
    """SHA1 checksum."""

    def __init__(self):
        super().__init__("sha1")


class SHA256(HashlibHasher):
    """SHA256 checksum."""

    def __init__(self):
        super().__init__("sha256")

    @classmethod
    def hash(
        cls,
        data: str | bytes,
        offset: Optional[int] = None,
        length: Optional[int] = None,
    ) -> bytes:
        """Gets sum of given data."""
        hasher = cls()
        hasher.update(
            data if isinstance(data, bytes) else data.encode(),
            offset,
            length,
        )
        return hasher.sum()


class MD5(HashlibHasher):
    """MD5 checksum."""

    def __init__(self):
        super().__init__("md5")

    @classmethod
    def hash(
        cls,
        data: bytes,
        offset: Optional[int] = None,
        length: Optional[int] = None,
    ) -> bytes:
        """Gets sum of given data."""
        hasher = cls()
        hasher.update(data, offset, length)
        return hasher.sum()


class Type(Enum):
    """Checksum algorithm type."""
    COMPOSITE = "COMPOSITE"
    FULL_OBJECT = "FULL_OBJECT"


class Algorithm(Enum):
    """Checksum algorithm."""
    CRC32 = "crc32"
    CRC32C = "crc32c"
    CRC64NVME = "crc64nvme"
    SHA1 = "sha1"
    SHA256 = "sha256"
    MD5 = "md5"

    def __str__(self) -> str:
        return self.value

    def header(self) -> str:
        """Gets headers for this algorithm."""
        return (
            "Content-MD5" if self == MD5 else f"x-amz-checksum-{self.value}"
        )

    def full_object_support(self) -> bool:
        """Checks whether this algorithm supports full object."""
        return self in {CRC32, CRC32C, CRC64NVME}

    def composite_support(self) -> bool:
        """Checks whether this algorithm supports composite."""
        return self in {CRC32, CRC32C, SHA1, SHA256}

    def validate(self, algo_type: Type):
        """Validates given algorithm type for this algorithm."""
        if not (
            (self.composite_support() and algo_type == Type.COMPOSITE)
            or (self.full_object_support() and algo_type == Type.FULL_OBJECT)
        ):
            raise ValueError(
                f"algorithm {self.name} does not support {algo_type.name} type",
            )

    def hasher(self):
        """Gets hasher for this algorithm."""
        if self == Algorithm.CRC32:
            return CRC32()
        if self == Algorithm.CRC32C:
            return CRC32C()
        if self == Algorithm.CRC64NVME:
            return CRC64NVME()
        if self == Algorithm.SHA1:
            return SHA1()
        if self == Algorithm.SHA256:
            return SHA256()
        if self == Algorithm.MD5:
            return MD5()
        return None


def new_hashers(
        algorithms: Optional[List[Algorithm]],
) -> Optional[Dict[Algorithm, "Hasher"]]:
    """Creates new hasher map for given algorithms."""
    hashers = {}
    if algorithms:
        for algo in algorithms:
            if algo and algo not in hashers:
                hashers[algo] = algo.hasher()
    return hashers if hashers else None


def update_hashers(
        hashers: Optional[Dict[Algorithm, "Hasher"]],
        data: bytes,
        length: int,
):
    """Updates hashers with given data and length."""
    if not hashers:
        return
    for hasher in hashers.values():
        hasher.update(data, 0, length)


def reset_hashers(hashers: Optional[Dict[Algorithm, "Hasher"]]):
    """Resets hashers."""
    if not hashers:
        return
    for hasher in hashers.values():
        hasher.reset()


def make_headers(
    hashers: Optional[Dict[Algorithm, "Hasher"]],
    add_content_sha256: bool,
    add_sha256_checksum: bool,
    algorithm_only: bool = False
) -> Dict[str, str]:
    """Makes headers for hashers.

    Args:
        hashers: Dictionary of algorithm to hasher instances
        add_content_sha256: Whether to add x-amz-content-sha256 header
        add_sha256_checksum: Whether to add SHA256 checksum header
        algorithm_only: If True, only include algorithm declaration header,
                       not checksum value headers
    """
    headers = {}
    if hashers:
        for algo, hasher in hashers.items():
            sum_bytes = hasher.sum()
            if algo == Algorithm.SHA256:
                if add_content_sha256:
                    headers["x-amz-content-sha256"] = hex_string(sum_bytes)
                if not add_sha256_checksum:
                    continue
            headers["x-amz-sdk-checksum-algorithm"] = str(algo)
            if not algorithm_only:
                headers[algo.header()] = base64_string(sum_bytes)
    return headers
